/* Time-stamp: <2009-11-14 18:22:51 poser> */
/*
 * Given a list of words, one per line, this program generates a list
 * of minimal pairs.
 *
 * Copyright (C) 1993-2009 William J. Poser (billposer@alum.mit.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * or go to the web page:  http://www.gnu.org/licenses/gpl.txt.
*/

#include "config.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#else
#define gettext(x) (x)
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <stddef.h>
#include <wchar.h>
#include "input.h"
#include "exitcode.h"
#include "unicode.h"
#define MINREPORTINTERVAL 1000000L
#define PBUFSIZE 256

#define TRUE 1
#define FALSE 0

int DebugP = 0;
char pgname[] = "minpair";
char version[]=PACKAGE_VERSION;
char compdate[]= __DATE__ " " __TIME__;
int VerboseP;			/* Chat while working? */
int plogfd;
unsigned long long int ReportInterval;

extern int ugetline(FILE *,wchar_t *, int);

static wchar_t *EmptyString = L"";
/* We keep the words on a linked list */

struct word{
  wchar_t *ctxt;		/* The word itself as processed for comparison */
  int clen;			/* The length of the word as processed, in characters */
  wchar_t *id;			/* The word's unique identifier */
  struct word *next;		/* Pointer to next item on linked list */
};


#define WORDPTR struct word *
#define NULLWORD ((struct word *) 0)
#define DEFWLENMAX 25 /* Default maximum characters in word */
#define DEFDSM '_'
/* #define DEFDSM 0x09593 */

static int wlmax;		/* Maximum word length */
static wchar_t SlotChar;	/* Character with which to mark difference site */

wchar_t pbuf[PBUFSIZE];
static wchar_t *tmpbuf;

void
ShowVersion(void){
  fprintf(stderr,"%s %s\n",pgname,version);
  fprintf(stderr,gettext("Compiled %s\n"),compdate);
  fprintf(stderr,"Copyright (C) 1993-2009 William J. Poser\n");
  fprintf(stderr,gettext("Released under the terms of version 3 of the GNU General Public License.\n"));
  fprintf(stderr,gettext("The GPL is available at http://www.gnu.org/copyleft/gpl.html.\n\n"));
}

void
PrintUsage(void)
{
  fprintf(stderr,
	  gettext("Generate an exhaustive list of minimal pairs from\na list of words, one per line.\n"));
  fprintf(stderr,gettext("By default, each word is followed by a tab and then an identifier\n"));
  fprintf(stderr,gettext("UTF-8 Unicode is accepted for input.\n"));
  fprintf(stderr,gettext("Input is read from a file or stdin if no filename is specified.\n"));
  fprintf(stderr,gettext("Output is written on stdout.\n"));
  fprintf(stderr,gettext("Usage: %s (<flags>) (<word list file>)\n"),pgname);
  fprintf(stderr,gettext("Flags:\n"));
  fprintf(stderr,gettext("\t-d <file containing definitions of up to 2048 multigraphs>\n"));
  fprintf(stderr,gettext("\t-h Print this help message.\n"));
  fprintf(stderr,gettext("\t-i Find pairs involving a single insertion/deletion\n"));
  fprintf(stderr,gettext("\t-l <maximum word length>\t(default = %d)\n"),DEFWLENMAX);
  fprintf(stderr,gettext("\t-m <difference site marker>\t(default = %c)\n"),DEFDSM);
  fprintf(stderr,gettext("\t-q quiet - do not chat while working\n"));
  fprintf(stderr,gettext("\t-s single column input - words lack identifiers\n"));
  fprintf(stderr,gettext("\t-t Find pairs involving a single transposition\n"));
  fprintf(stderr,gettext("\t-v Print version information.\n"));
}

FILE *
OpenFile(char *file,char *mode,char *pgname)
{
  FILE *fp;
  extern FILE *fopen();
   
  if((fp = fopen(file,mode)) != NULL) return fp;
  else{
    switch(*mode){
    case 'r': 
      fprintf(stderr,gettext("%s: can't open file %s for reading.\n"),pgname,file);
      break;
    case 'w':
      fprintf(stderr,gettext("%s: can't open file %s for writing.\n"),pgname,file);
      break;
    default:
      fprintf(stderr,gettext("%s: can't open file %s for appending.\n"),pgname,file);
    }
    exit(OPENERROR);
  }
  /* NOTREACHED */
}

/*
 * Copy a string and return a pointer to the new storage allocated
 * to contain it. This function differs from strcpy(3) in that it
 * allocates the necessary storage instead of just copying into
 * storage provided by the caller.
 */

char *
copy_string(char *string)
{
   char *new;
 
   if( (new = (char *)malloc( (size_t) (strlen(string) + 1) ) ) != (char *) 0){
      strcpy(new,string);
   }
   else fprintf(stderr,gettext("copy_string: malloc error\n"));

   return(new);
}

/*
 * We handle multigraphs by mapping them to unused single codepoints so that
 * they will compare as single segments. On output, we expand them again into
 * the original string. Since the Surrogate ranges are large enough
 * and are unused in input originating in UTF-8, we use these ranges for the
 * internal codes. The bulk of the work is involved in mapping multigraphs
 * to internal codes on input, but we also generate an array that maps
 * internal codes back to the original strings so that we can expand
 * the internal representation back to the orginal sequences on output.
 */

#define FIRSTMULTIGRAPHCODEPOINT 0xD800	/* First codepoint in High Surrogate range */
#define MAXMULTIGRAPHCODEPOINT   0xDFFF /* Last codepoint in Low Surrogate range */
#define MAXMULTIGRAPHS (1+MAXMULTIGRAPHCODEPOINT - FIRSTMULTIGRAPHCODEPOINT)
#define MULTIGRAPHBUFFERSIZE 64


/* The map table, for mapping sequences to codes */
struct mapentry{
   wchar_t *str;		/* The multigraph */
   wchar_t code;		/* The codepoint to use internally */
   size_t len;			/* Length of str */
};
struct mapentry *MapTable[MAXMULTIGRAPHS];

int MapTableEntries;
wchar_t NextCode;		/* Next code available for multigraph */
wchar_t *MTable[MAXMULTIGRAPHS]; /* For expanding again on output */

void
AddStringToMapTable(wchar_t *s)
{
   struct mapentry *new;

   if(NextCode > MAXMULTIGRAPHCODEPOINT){
      fprintf(stderr,gettext("%s:Too many multigraphs defined.\n"),pgname);
      exit(LIMITEXCEEDED);
   }
   
   /* Allocate a new map table entry */
   new = (struct mapentry *) malloc(sizeof(struct mapentry));
   if(new == (struct mapentry *) 0){
      fprintf(stderr,gettext("%s: out of memory - map table entry\n"),pgname);
      exit(OUTOFMEMORY);
   }
   new->len = wcslen(s);
   new->str = (wchar_t *) malloc((new->len + 1) * sizeof(wchar_t));
   if(new->str == NULL){
     fprintf(stderr,gettext("%s:out of memory - map table string\n"),pgname);
     exit(OUTOFMEMORY);
   }
   (void)wcscpy(new->str,s);
   new->code = NextCode;
   /* We also need a mapping of internal code to the original string to use on output  */
   MTable[NextCode - FIRSTMULTIGRAPHCODEPOINT] = new->str;
   NextCode++;
   MapTable[MapTableEntries] = new;
   MapTableEntries+=1;
}

/* Sort the map table so that sequence length is decreasing */

void
SortMapTable(void)
{
   int gap;
   int i;
   int j;
   struct mapentry *temp;
   
   for(gap = MapTableEntries/2; gap > 0; gap /= 2){
      for(i = gap; i < MapTableEntries; i++){
	 for(j = i-gap; j>= 0; j -= gap){
	    if(MapTable[j]->len >= MapTable[j+gap]->len) break;
	    else{
	       temp = MapTable[j];
	       MapTable[j] = MapTable[j+gap];
	       MapTable[j+gap] = temp;
	    }
	 }
      }
   }
}

void
ReadMultigraphDefinitions(char *file)
{
   FILE *fp;
   int LineLength;
   wchar_t buf[MULTIGRAPHBUFFERSIZE];
   wchar_t *Terminators=L"\x0020\x0009";
   wchar_t *wcstokStore;
   wchar_t *FStart;
   extern short UGetLineDone;
   
   fp = OpenFile(file,"r",pgname);
   UGetLineDone =0; 
   while( (LineLength = ugetline(fp,buf,MULTIGRAPHBUFFERSIZE)) != ENDOFINPUT){
     if(LineLength == 1) continue;
     if(LineLength < 1) break;
     for (FStart = wcstok(buf,Terminators,&wcstokStore);
	  FStart != NULL;
	  FStart = wcstok(NULL,Terminators,&wcstokStore)) {
       AddStringToMapTable(FStart);
     }
   }
   fclose(fp);
   SortMapTable();
}

/*
 * Map multigraphs to internal codes.
 * This presupposes that the map table has been sorted by SortMapTable()
 * into decreasing order of length.
 *	 
 *	Start at the beginning of the input string.
 * While there are still characters in the input string
 * 	Set the index into the MapTable to zero
 *		Work through MapTable, looking for prefixes of the input string.
 * 	If we find a prefix
 *			copy the corresponding code to the output
 *			advance the input pointer by the length of the prefix
 *			try again
 *		If we don't find a prefix
 *			copy the current character to the output
 *			advance the input pointer one character
 *			try again
 * Else we're done mapping this string
 */	 

int
MapMultigraphs(wchar_t *in, wchar_t *out)
{
   int mte;
   int j = 0;
   int FoundP;
   
   if(MapTableEntries == 0){		/* Don't bother with fancy mapping */
      while(*in != '\0') out[j++] = *in++;
      out[j]= (wchar_t) '\0';
      return(j);
   }
   
   FoundP = FALSE;
   while(*in != (wchar_t) '\0'){
      mte = 0;
      FoundP = FALSE;
      while(mte < MapTableEntries){
	 if(wcsncmp(in,MapTable[mte]->str,MapTable[mte]->len) == 0){
	    FoundP = TRUE;
	    out[j++] = MapTable[mte]->code;
	    in+=MapTable[mte]->len;
	    break;
	 }
	 ++mte;
      }
      if(!FoundP) out[j++] = *in++;
   }
   out[j]= (wchar_t) '\0';
   return(j);
}

/* Allocate a new word node and copy new word and id into it */
WORDPTR
NewWordWithID(wchar_t *inbuf, int len, unsigned long lineno)
{
  WORDPTR new;
  size_t idlen;
  size_t wdlen;
  wchar_t *tbk;			/* Pointer to first tab in inbuf */
  wchar_t *id;

  /* Create new word node */
  new = (WORDPTR) malloc (sizeof (struct word));
  if(new == NULLWORD){
    fprintf(stderr,gettext("%s: out of memory.\n"),pgname);
    exit(OUTOFMEMORY);
  }
  new->next = NULLWORD;

  /* Parse line into word and id */
  tbk = wcschr(inbuf,L'\t');
  if(tbk == NULL) {
    fprintf(stderr,gettext("ill-formed single-column input at line %lu\n"),lineno);
    exit(BADRECORD);
  }
  *tbk=L'\0';			/* Split line into two strings. */

  /* Handle word */
  wdlen = (size_t)(tbk - inbuf);
  new->ctxt = (wchar_t *) malloc((wdlen +1) * sizeof(wchar_t));
  new->clen = MapMultigraphs(inbuf,tmpbuf);
  if(new->ctxt == (wchar_t *) 0){
    fprintf(stderr,gettext("%s: out of memory.\n"),pgname);
    exit(OUTOFMEMORY);
  }
  (void) wcscpy(new->ctxt,tmpbuf);

  /* Handle id */
  id = tbk+1;
  idlen = wcslen(id);
  new->id = (wchar_t *) malloc((idlen +1) * sizeof(wchar_t));
  if(new->id == (wchar_t *) 0){
    fprintf(stderr,gettext("%s: out of memory.\n"),pgname);
    exit(OUTOFMEMORY);
  }
  (void) wcscpy(new->id,id);
  return(new);
}

/* Allocate a new word node and copy new word into it */

WORDPTR
NewWordAlone(wchar_t *inbuf, int len, unsigned long lineno)
{
  WORDPTR new;
   
  new = (WORDPTR) malloc (sizeof (struct word));
  if(new == NULLWORD){
    fprintf(stderr,gettext("%s: out of memory.\n"),pgname);
    exit(OUTOFMEMORY);
  }
  new->next = NULLWORD;
  new->ctxt = (wchar_t *) malloc((len +1) * sizeof(wchar_t));
  new->clen = MapMultigraphs(inbuf,tmpbuf);
  if(new->ctxt == (wchar_t *) 0){
    fprintf(stderr,gettext("%s: out of memory.\n"),pgname);
    exit(OUTOFMEMORY);
  }
  (void) wcscpy(new->ctxt,tmpbuf);
  new->id = EmptyString;
  return(new);
}

/*
 * Read the input file line by line and copy each into a new
 *	node in the linked list.
 */

unsigned long
LoadWords(FILE *fp,WORDPTR wptr, int IDP)
{
  unsigned long cnt;
  unsigned long lineno;
  int BlankLines;
  int OverlongLines;
  int wlen; /* Length of word in characters */
  wchar_t *inbuf;
  WORDPTR new;
  WORDPTR NewWord(wchar_t *,int);
  extern short UGetLineDone;

  extern void putu8s(wchar_t *, FILE *);
  
  inbuf = (wchar_t *) malloc((wlmax +1) * sizeof(wchar_t));
  if(inbuf == NULL){
    fprintf(stderr,gettext("%s: out of memory.\n"),pgname);
    exit(OUTOFMEMORY);
  }
  tmpbuf = (wchar_t *) malloc((wlmax +1) * sizeof(wchar_t));
  if(tmpbuf == NULL){
    fprintf(stderr,gettext("%s: out of memory.\n"),pgname);
    exit(OUTOFMEMORY);
  }
  lineno = 0L;
  cnt = 0L;
  BlankLines = 0;
  OverlongLines = 0;
  UGetLineDone = 0; 
  while( (wlen = ugetline(fp,inbuf,wlmax)) != ENDOFINPUT){
    lineno++;
    switch(wlen) {
    case 0:
      if(VerboseP) {
	fprintf(stderr,gettext("Blank line at line    %5lu.\n"),lineno);
      }
      ++BlankLines;
      break;
    case BUFOVERFLOW:
      if(VerboseP) {
	fprintf(stderr,gettext("Overlong line at line %5lu.\n"),lineno);
	//      swprintf(pbuf,PBUFSIZE-1,gettext("Word at line %5lu beginning [%ls] too long - skipped.\n"),
      //	       lineno,inbuf);
    //      putu8s(pbuf,stderr);
      }
      ++OverlongLines;
      break;
    default:   
      cnt+=1L;
      new = (IDP ? NewWordWithID(inbuf,wlen,lineno) : NewWordAlone(inbuf,wlen,lineno));
      wptr->next = new;
      wptr = new;
    }
  }
  if (VerboseP) {
    if(BlankLines == 1) fprintf(stderr,gettext("1 blank line skipped.\n"));
    if(BlankLines > 1) fprintf(stderr,gettext("%d blank lines skipped.\n"),BlankLines);
    if(OverlongLines == 1) fprintf(stderr,gettext("1 overlong line skipped.\n"));
    if(OverlongLines > 1) fprintf(stderr,gettext("%d overlong lines skipped.\n"),OverlongLines);
  }
  free( (void *) inbuf);
  free( (void *) tmpbuf);
  return(cnt);
}


void
ReportProgress(unsigned long long int comps,unsigned long long int total)
{
  int percent;	
  long double dcomps;
  long double dtotal;
  char plogbuf[5];

  dcomps = (long double) comps;
  dtotal = (long double) total;
  percent =  100 * (dcomps/dtotal);
  if(VerboseP) {
    fprintf(stderr,"\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b%'15llu (%2d%%)",
	    comps,(int)percent);
  }
  sprintf(plogbuf,"%3d\n",percent);
  write(plogfd,plogbuf,4);
  fsync(plogfd);
}

/* 
 * Convert a null terminated UTF-32 string to UTF-8 and write it out.
 * Whenever a character in the High Surrogate range is encountered,
 * expand it to the original multigraph.
 */

void
ExpandPutu8s(wchar_t *str,FILE *fp)
{
  wchar_t *ptr;
  wchar_t c;
  wchar_t index;
  extern void fputu8(wchar_t, FILE *);
  extern void putu8s(wchar_t *, FILE *);

  ptr = str;
  while( (c = *ptr++) != L'\0'){
    if( (c >= FIRSTMULTIGRAPHCODEPOINT) && (c <= MAXMULTIGRAPHCODEPOINT)){
      index = c-FIRSTMULTIGRAPHCODEPOINT;
      if(MTable[index] != NULL) putu8s(MTable[index],fp); 
    }
    else fputu8(c,fp);
  }
}

int
Hamming(struct word *A, struct word *B)
{
  wchar_t *a;
  wchar_t *b;
  int difs = 0;
  int index = 0;
  int difind;
  wchar_t ADifChar;
  wchar_t FirstChar;wchar_t SecondChar;wchar_t TempChar;
  wchar_t *FirstWord;wchar_t *SecondWord;wchar_t *TempWord;
  wchar_t *FirstID;wchar_t *SecondID;wchar_t *TempID;
  
  a = A->ctxt;
  b = B->ctxt;
  while(*a != '\0'){
    if(*a != *b){
      difs+=1;
      difind = index;
      if(difs > 1) return (0);
    }
    ++a;++b;++index;
  }
  if(difs == 1){
    FirstChar = A->ctxt[difind];
    SecondChar = B->ctxt[difind];
    FirstWord = A->ctxt;
    SecondWord = B->ctxt;
    FirstID = A->id;
    SecondID = B->id;
    if(FirstChar > SecondChar){ /* Swap to reorder */
      TempChar = FirstChar;FirstChar = SecondChar;SecondChar = TempChar;
      TempWord = FirstWord;FirstWord = SecondWord;SecondWord = TempWord;
      TempID = FirstID;FirstID = SecondID;SecondID = TempID;
    }
    swprintf(pbuf,PBUFSIZE-1,L"%lc\t%lc\t%ls\t%ls\t%ls\t%ls\t",
	     FirstChar,SecondChar,FirstWord,FirstID,SecondWord,SecondID);
    ExpandPutu8s(pbuf,stdout);
    ADifChar = A->ctxt[difind];
    A->ctxt[difind] = '\0'; /* Split into two parts */
    swprintf(pbuf,PBUFSIZE-1,L"%ls%lc%ls\n",A->ctxt,SlotChar,A->ctxt+difind+1);
    ExpandPutu8s(pbuf,stdout);
    A->ctxt[difind] = ADifChar; /* Restore */
    return(1);
  }
  return(0);
}

/*
 * Given two words known to differ in length by exactly one character,
 * determine whether they differ by a single deletion from the longer
 * word. If so, report a minimal pair. 	 
 */

int
InDel(struct word *A, struct word *B)
{
  wchar_t *Shorter;
  wchar_t *Longer;
  int ShorterLen;
  int FirstPartLen;
  int SecondPartLen;
  wchar_t DifChar;
  int FoundP;
  int LenA, LenB;
  wchar_t *ShorterID;
  wchar_t *LongerID;
  
  LenA = A->clen;
  LenB = B->clen;

  FoundP = FALSE;
  if(LenA < LenB){
    Shorter = A->ctxt;
    ShorterLen = LenA;
    Longer = B->ctxt;
    ShorterID = A->id;
    LongerID = B->id;
  }
  else{
    Shorter = B->ctxt;
    ShorterLen = LenB;
    Longer = A->ctxt;
    ShorterID = B->id;
    LongerID = A->id;
  }
  if(wcsncmp(Shorter,Longer,ShorterLen) == 0){ /* Prefix */
    swprintf(pbuf,PBUFSIZE-1,L"0\t%lc\t%ls\t%ls\t%ls\t%ls\t%ls%lc\n",
	     Longer[ShorterLen],
	     Shorter,
	     ShorterID,
	     Longer,
	     LongerID,
	     Shorter,SlotChar);
    ExpandPutu8s(pbuf,stdout);
    FoundP = TRUE;
  }
  else if(wcsncmp(Shorter,Longer+1,ShorterLen) == 0){ /* Suffix */
    swprintf(pbuf,PBUFSIZE-1,L"0\t%lc\t%ls\t%ls\t%ls\t%ls\t%lc%ls\n",
	     Longer[0],
	     Shorter,
	     ShorterID,
	     Longer,
	     LongerID,
	     SlotChar,Shorter);
    ExpandPutu8s(pbuf,stdout);
    FoundP = TRUE;
   }
  else{
    for(FirstPartLen = 1; FirstPartLen < ShorterLen; FirstPartLen++){
      SecondPartLen = ShorterLen - FirstPartLen;
      if((wcsncmp(Shorter,Longer,FirstPartLen) == 0) &&
	 (wcsncmp(Shorter+FirstPartLen,Longer+FirstPartLen+1,SecondPartLen)
	  == 0)){
	DifChar = Longer[FirstPartLen];
	swprintf(pbuf,PBUFSIZE-1,L"0\t%lc\t%ls\t%ls\t",
		 DifChar,
		 Shorter,
		 ShorterID,
		 Longer,
		 LongerID);
	ExpandPutu8s(pbuf,stdout);
	Longer[FirstPartLen] = '\0'; /* Split pieces */
	swprintf(pbuf,PBUFSIZE-1,L"%ls%lc%ls\n",Longer,SlotChar,Longer+FirstPartLen+1);
	ExpandPutu8s(pbuf,stdout);
	Longer[FirstPartLen] = DifChar; /* Restore */
	FoundP = TRUE;
	break;
      }
    }
  }
  return(FoundP);
}


int
Transpositions(struct word *A, struct word *B)
{
  int i;
  int FoundP;
  int len;

  wchar_t SavedChar;
  wchar_t FirstChar;wchar_t SecondChar;wchar_t TempChar;
  wchar_t *FirstWord;wchar_t *SecondWord;wchar_t *TempWord;
  wchar_t *FirstID;wchar_t *SecondID;wchar_t *TempID;
   
  FoundP = FALSE;
  len = A->clen;
   
  /* Transposition is prefix*/
  if((A->ctxt[0] == B->ctxt[1]) && (A->ctxt[1] == B->ctxt[0]) && (wcscmp(A->ctxt+2,B->ctxt+2) == 0)){
    FirstChar = A->ctxt[0];
    SecondChar = A->ctxt[1];
    FirstWord = A->ctxt;
    SecondWord = B->ctxt;
    FirstID = A->id;
    SecondID = B->id;
    if(FirstChar > SecondChar){ /* Swap to reorder */
      TempChar = FirstChar;FirstChar = SecondChar;SecondChar = TempChar;
      TempWord = FirstWord;FirstWord = SecondWord;SecondWord = TempWord;
      TempID = FirstID;FirstID = SecondID;SecondID = TempID;
    }
    swprintf(pbuf,PBUFSIZE-1,L"%lc\t%lc\t%ls\t%ls\t%ls\t%ls\t%lc%lc%ls\n",
	     FirstChar,SecondChar,
	     FirstWord,
	     FirstID,
	     SecondWord,
	     SecondID,
	     SlotChar,SlotChar,FirstWord+2);
    ExpandPutu8s(pbuf,stdout);
    FoundP = TRUE;
  }
  /* Transposition is suffix */
  else if((A->ctxt[len-1] == B->ctxt[len-2]) && (A->ctxt[len-2] == B->ctxt[len-1])
	  && (wcsncmp(A->ctxt,B->ctxt,len-2) == 0)){
    FirstChar = A->ctxt[len-1];
    SecondChar = B->ctxt[len-1];
    FirstWord = A->ctxt;
    SecondWord = B->ctxt;
    FirstID = A->id;
    SecondID = B->id;
    if(FirstChar > SecondChar){ /* Swap to reorder */
      TempChar = FirstChar;FirstChar = SecondChar;SecondChar = TempChar;
      TempWord = FirstWord;FirstWord = SecondWord;SecondWord = TempWord;
      TempID = FirstID;FirstID = SecondID;SecondID = TempID;
    }
    swprintf(pbuf,PBUFSIZE-1,"%lc\t%lc\t%ls\t%ls\t%ls\t%ls\t",
	     FirstChar,SecondChar,
	     A->ctxt,
	     A->id,
	     B->ctxt,
	     B->id);
    ExpandPutu8s(pbuf,stdout);
    SavedChar = A->ctxt[len-2];
    A->ctxt[len-2] = '\0';
    swprintf(pbuf,PBUFSIZE-1,L"%ls%lc%lc\n",A->ctxt,SlotChar,SlotChar);
    ExpandPutu8s(pbuf,stdout);
    A->ctxt[len-2] = SavedChar;
    FoundP = TRUE;
  }
  else{ /* Transposition is not at either edge */
    for(i = 1; i < len-2; i++){
      if((A->ctxt[i] == B->ctxt[i+1]) && (A->ctxt[i+1] == B->ctxt[i]) &&
	 (wcsncmp(A->ctxt,B->ctxt,i) == 0) &&
	 (wcscmp(A->ctxt+i+2,B->ctxt+i+2) == 0)){
	FirstChar = A->ctxt[i];
	SecondChar = B->ctxt[i];
	FirstWord = A->ctxt;
	SecondWord = B->ctxt;
	FirstID = A->id;
	SecondID = B->id;
	if(FirstChar > SecondChar){ /* Swap to reorder */
	  TempChar = FirstChar;FirstChar = SecondChar;SecondChar = TempChar;
	  TempWord = FirstWord;FirstWord = SecondWord;SecondWord = TempWord;
	  TempID = FirstID;FirstID = SecondID;SecondID = TempID;
	}
	swprintf(pbuf,PBUFSIZE-1,L"%lc\t%lc\t%ls\t%ls\t",
		 FirstChar,SecondChar,
		 A->ctxt,
		 A->id,
		 B->ctxt,
		 B->id);
	ExpandPutu8s(pbuf,stdout);
	SavedChar = A->ctxt[i];
	A->ctxt[i] = '\0';
	swprintf(pbuf,PBUFSIZE-1,L"%ls%lc%lc%ls\n",A->ctxt,SlotChar,SlotChar,A->ctxt+i+2);
	ExpandPutu8s(pbuf,stdout);
	A->ctxt[i] = SavedChar;
	FoundP = TRUE;
      }
    }
  }
  return(FoundP);
}

/*
 * Do the comparisons.
 *
 * Notice that we could make this faster when doing transpositions
 * by calling Transpositions() from within Hamming, since a transpostion
 * is only possible if the Hamming distance is 2. However, this makes
 * it run slower when not doing transpositions and makes the program
 * structure ugly, so I haven't done this.
 */


long
CheckWords(WORDPTR WordList, unsigned long long total,int DoIndelsP,
	   int DoTransP)
{
  WORDPTR A;
  WORDPTR B;
  unsigned long long Comparisons;	/* Number of comparisons made thus far */
  long MinPairs;		/* Number of minimal pairs found */
  int LengthA;
  int LengthB;
  int DeltaLen;
   
  Comparisons = 0L;
  MinPairs = 0L;
   
  A = WordList->next; /* Skip the head since it has a null txt field */
  while(A != NULLWORD){
    B = A->next;
    while(B != NULLWORD){
      if(++Comparisons % ReportInterval == 0){
	ReportProgress(Comparisons,total);
      }
      LengthA = A->clen;
      LengthB = B->clen;
      DeltaLen = abs(LengthA - LengthB);
      if(DeltaLen == 0){
	if(Hamming(A,B)) ++MinPairs;
	else if(DoTransP){
	  if(Transpositions(A,B)) ++MinPairs;
	}
      }
      else if(DoIndelsP && (DeltaLen == 1)){
	if(InDel(A,B)) ++MinPairs;
      }
      B = B->next;
    }
    A = A->next;
  }
  ReportProgress(Comparisons,total);
  fprintf(stderr,"\n");
  return(MinPairs);
}

int
main(int ac, char **av)
{
  FILE *infp;			/* Pointer to file contain word list */
  int c;			/* Option returned by getopt() */
  unsigned long WordCount;		/* Number of words on list */
  unsigned long long int TotalComps;	/* Total comparisons made */
  long MinPairs;		/* Number of minimal pairs found */
  struct word WordListHead;	/* Head of linked list of words */
  int DoIndelsP;		/* Find pairs involving indels */
  int DoTransP;		     /* Find pairs involving transpositions */
  char *DefinitionFileName = NULL;	/* Name of file containing multigraph definitions */
  char *tmpptr;			/* Temporary pointer */
  int i;
  int IDP = 1;			/* Are input words accompanied by identifiers? */
   
   
  /* Defaults */
   
   wlmax = DEFWLENMAX;
   SlotChar = DEFDSM;
   DoIndelsP = FALSE;
   DoTransP = FALSE;
   VerboseP = TRUE;


#ifdef HAVE_SETLOCALE
   tmpptr = setlocale(LC_ALL,"");
   if (tmpptr && DebugP) fprintf(stderr,"Locale:\t%s\n",tmpptr);
#endif
#ifdef HAVE_LIBINTL_H
   tmpptr = bindtextdomain (PACKAGE,LOCALEDIR);
   if (tmpptr && DebugP) fprintf(stderr,"Base directory:\t%s\n",tmpptr);
   textdomain (PACKAGE);
#endif
   
   /* Command line options */
   
   while( (c = getopt(ac,av,"d:hil:m:qstv")) != EOF){
      switch(c){
      case 'd':
	DefinitionFileName = copy_string(optarg);
	break;
      case 'h':
	PrintUsage();
	exit(2);
	break;
       case 'i':
	 DoIndelsP = TRUE;
	 break;
       case 'l':
	 wlmax = atoi(optarg);
	 break;
       case 'm':
	 if(mbtowc(&SlotChar,optarg,6) < 1){
	   fprintf(stderr,gettext("Invalid slot character specification.\n"));
	   exit(3);
	 }
	 break;
      case 'q':
	VerboseP = 0;
	break;
       case  's':
	 IDP = FALSE;
	 break;
       case 't':
	 DoTransP = TRUE;
	 break;
      case 'v':
	ShowVersion();
	exit(2);
       default:			/* NOTREACHED */
	 PrintUsage();
	 exit(INFO);
	 break;
      }
   }
   
   /* Open progress file */
   plogfd = open("minpair.plog",O_CREAT | O_WRONLY | O_TRUNC|O_SYNC,00644);
   if(plogfd < 0) {
     fprintf(stderr,"Cannot open plog file.\n");
     exit(OPENERROR);
   }

   /* Report configuration used */

   if(VerboseP){
     if(DoIndelsP && DoTransP){
       fprintf(stderr,gettext("Searching for insertion/deletion pairs and tranpositions\nas well as pairs with same order and length.\n"));
     }
     else if(DoIndelsP){
       fprintf(stderr,gettext("Searching for insertion/deletion pairs as well as pairs\nwith same order and length.\n"));      
     }
     else if(DoTransP){
       fprintf(stderr,gettext("Searching for transpositions as well as pairs with same\norder and length.\n"));
     }
     else fprintf(stderr,gettext("Searching only for pairs with same order and length.\n"));
   }

   /* Initialize MTable for safety */
   for(i=0; i < MAXMULTIGRAPHS; i++) MTable[i] = NULL;

   /* Get multigraph definitions */
   NextCode = FIRSTMULTIGRAPHCODEPOINT;
   MapTableEntries = 0;
   if(DefinitionFileName != NULL) ReadMultigraphDefinitions(DefinitionFileName);
   if(VerboseP) {
     if(MapTableEntries > 0)fprintf(stderr,gettext("%d multigraphs defined.\n"),MapTableEntries);
     else fprintf(stderr,gettext("No multigraphs defined.\n"));
   }

   /* Load word list */
   
   if(optind < ac) infp = OpenFile(av[optind],"r",pgname);
   else infp = stdin;
   WordCount = LoadWords(infp,&WordListHead,IDP);
   fclose(infp);
   if(WordCount < 2L){
      fprintf(stderr,gettext("%s: there's no point in comparing fewer than two words.\n"),pgname);
      exit(OTHERERROR);
   }
   if(VerboseP) fprintf(stderr,gettext("Read %'lu words.\n"),WordCount);

   /* Find pairs */
   
   TotalComps = (unsigned long long int) WordCount;
   TotalComps = TotalComps * (TotalComps - 1ULL)/2ULL;
   ReportInterval = TotalComps / 50L;
   if (ReportInterval < MINREPORTINTERVAL) ReportInterval = TotalComps / 25L;
   if (ReportInterval < MINREPORTINTERVAL) ReportInterval = MINREPORTINTERVAL;
   if(VerboseP) {
     if(TotalComps > ReportInterval){
       fprintf(stderr,gettext("Comparisons made, out of %'llu: %'15llu (%2d%%)"),TotalComps,0ULL,0);
     }
   }
   MinPairs = CheckWords(&WordListHead,TotalComps,DoIndelsP,DoTransP);
   
   /* Clean up and exit */
   
   if(VerboseP) { 
     if(MinPairs == 1) fprintf(stderr,gettext("1 minimal pair found.\n"));
     else fprintf(stderr,gettext("%ld minimal pairs found.\n"),MinPairs);
   }
   close(plogfd);
   exit(SUCCESS);
}

